/* cofflayout.cpp */

#include "cofflayout.h"
#include <cstring> // strnlen

const char* formatMachineType(uint16 machine) {
	switch(machine) {
		case 0: return "any";
		case 0x8664: return "x86_64";
		case MACHINE_TYPE_i386: return "i386";
		case 0x1c0: return "ARMle";
		case 0x1c4: return "ARMv7";
		case 0xaa64: return "AArch64";
		case 0x200: return "Itanium";
		default: return "other/unknown";
	}
}

const char* formatStorageClass(uchar storageClass) {
	switch (storageClass) {
		case 0:   return "null";
		case 2:   return "extern";
		case 3:   return "static";
		case 101: return "func";
		case 103: return "file";
		default:  return "other";
	}
}

const char* formatI386RelocationType(uint16 relocationType) {
	switch(relocationType) {
		case 0:    return "absolute";
		case RELOCATION_TYPE_DIR32:   return "DIR32 (VA)";
		case RELOCATION_TYPE_DIR32NB: return "DIR32NB (RVA)";
		case RELOCATION_TYPE_REL32:   return "REL32";
		default:   return "other";
	}
}

std::string coffSymbol_getName(coffSymbol* cs, uint32* stringTableSizePtr) {
	if (cs->name.offset.zero == 0) {
		char* str = (char*)(stringTableSizePtr)+cs->name.offset.stringTableOffset;
		return str;

	} else {
		return std::string(cs->name.shortName, strnlen(cs->name.shortName, 8));
	}
}
