/* mld/defs.h */

#ifndef MLD_COFFLAYOUT_H
#define MLD_COFFLAYOUT_H

typedef signed short sint16;
typedef unsigned short uint16;
typedef unsigned int uint32;
typedef unsigned char uchar;

struct dosHeader {
	uchar magic[2];
	uchar unused[0x3c - 2];
	uint32 pe_offset;
};

struct coffHeader {
	uint16 machine;
	uint16 numSections;
	uint32 timeStamp;
	uint32 ptrSymbolTable; // zero/deprecated
	uint32 numSymbols;     // zero/deprecated
	uint16 sizeOfOptionalHeader;
	uint16 characteristics;
};

struct dataDirectory {
	uint32 virtualAddr;
	uint32 size;
};

struct optionalHeaderStandard {
	uint16 magic;
	uchar linkerVersionMajor;
	uchar linkerVersionMinor;
	uint32 sizeOfText;
	uint32 sizeOfData;
	uint32 sizeOfBSS;
	uint32 addressOfEntryPoint;
	uint32 baseOfCode;
	uint32 baseOfData; // absent in PE32+
};

struct optionalHeaderWin32 {
	uint32 imageBase;
	uint32 alignSections;
	uint32 alignFile;
	uint16 osVersionMajor; // minimum OS version
	uint16 osVersionMinor;
	uint16 imageVersionMajor;
	uint16 imageVersionMinor;
	uint16 subsystemVersionMajor;
	uint16 subsystemVersionMinor;
	uint32 win32Version; // must be zero
	uint32 sizeOfImage;
	uint32 sizeOfHeaders;
	uint32 checksum; // optional
	uint16 subsystem;
	uint16 dllCharacteristics;
	uint32 sizeOfStackReserve;
	uint32 sizeOfStackCommit;
	uint32 sizeOfHeapReserve;
	uint32 sizeOfHeapCommit;
	uint32 loaderFlags; // must be zero
	uint32 numRvaAndSizes;
};

struct optionalHeaderDirectory {
	dataDirectory
		exportTable,
		importTable,
		resourceTable,
		exceptionTable,
		certificateTable,
		baseRelocationTable,
		debug,
		architecture, // must be zero
		globalPtr,
		tlsTable,
		loadConfigTable,
		boundImport,
		importAddressTable,
		delayImportDescriptor,
		CLRRuntimeHeader,
		reserved; // must be zero
};

struct optHeader {
	optionalHeaderStandard s;
	optionalHeaderWin32 w;
	optionalHeaderDirectory d;
};

struct sectionHeader {
	char name[8];
	uint32 virtualSize;
	uint32 virtualAddress;
	uint32 sizeOfRawData;
	uint32 fpRawData;
	uint32 fpRelocations;
	uint32 fpLineNumbers; // zero, deprecated
	uint16 numRelocations;
	uint16 numLineNumbers; // zero, deprecated
	uint32 characteristics;
};

struct relocation {
	uint32 virtualAddress;
	uint32 symbolTableIndex;
	uint16 type;
};

#define SIZEOF_RELOCATION 10

struct coffSymbol {
	union {
		char shortName[8];
		struct {
			uint32 zero;
			uint32 stringTableOffset;
		} offset;
	} name;

	uint32 value;
	sint16 sectionNumber; // signed
	uint16 type;
	uchar storageClass;
	uchar numAuxSymbols;
};

#define SIZEOF_COFFSYMBOL 18

#define MACHINE_TYPE_i386 0x14c

#define RELOCATION_TYPE_DIR32 0x6
#define RELOCATION_TYPE_DIR32NB 0x7
#define RELOCATION_TYPE_REL32 0x14

enum COFF_CHARACTERISTICS {
	IMAGE_FILE_RELOCS_STRIPPED = 0x0001,
	IMAGE_FILE_EXECUTABLE_IMAGE = 0x0002,
	IMAGE_FILE_LINE_NUMS_STRIPPED = 0x0004,
	IMAGE_FILE_LOCAL_SYMS_STRIPPED = 0x0008,
	IMAGE_FILE_AGGRESSIVE_WS_TRIM = 0x0010,
	IMAGE_FILE_LARGE_ADDRESS_AWARE = 0x0020,
	IMAGE_FILE_BYTES_REVERSED_LO = 0x0080,
	IMAGE_FILE_32BIT_MACHINE = 0x0100,
	IMAGE_FILE_DEBUG_STRIPPED = 0x0200,
	IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP = 0x0400,
	IMAGE_FILE_NET_RUN_FROM_SWAP = 0x0800,
	IMAGE_FILE_SYSTEM = 0x1000,
	IMAGE_FILE_DLL = 0x2000,
	IMAGE_FILE_UP_SYSTEM_ONLY = 0x4000,
	IMAGE_FILE_BYTES_REVERSED_HI = 0x8000
};



/* FUNCTION PROTOTYPES IN COFFLAYOUT.CPP */

const char* formatMachineType(uint16 machine);
const char* formatStorageClass(uchar storageClass);
const char* formatI386RelocationType(uint16 relocationType);

#include <string>
std::string coffSymbol_getName(coffSymbol* cs, uint32* stringTableSizePtr);

#endif
