#include "linker.h"
#include <cstring> // strncpy

sectionOffsetInfo::sectionOffsetInfo(linkOptions& lo, sectiondata_t& sectionData) {
	totalImageLengthFile = 0;
	totalImageLengthMem = lo.sectionAlign; // ASSUMING lo.sectionAlign > size of headers(!)

	for (sectiondata_t::iterator i = sectionData.begin(), e = sectionData.end(); i != e; ++i) {

		uint32 len = i->second.length();

		uint32 align = lo.fileAlign - (len % lo.fileAlign);
		if (align == lo.fileAlign) align = 0;
		if (len == 0) align = lo.fileAlign;
		sectionAlignmentsFile[i->first] = align;

		uint32 sectalign = lo.sectionAlign - (len % lo.sectionAlign);
		if (sectalign == lo.sectionAlign) sectalign = 0;
		if (len == 0) sectalign = lo.sectionAlign;
		sectionAlignmentsMem[i->first] = sectalign;

		sectionMemPos[i->first] = totalImageLengthMem;

		totalImageLengthFile += len + align;
		totalImageLengthMem += len + sectalign;
	}
}

int linker_write_file(
	linkOptions& lo,
	sectiondata_t& sectionData,
	sectioncharac_t& sectionCharacteristics,
	uint32 entry_rva
) {

	sectionOffsetInfo info = sectionOffsetInfo(lo, sectionData);

	//
	// Output linked file
	//

	if (lo.output == NULL) {
		fputs("No output file specified", stderr);
		return 1;
	}

	// MZ stub / PE header

	const unsigned char mzstub[] = {
		0x4D, 0x5A, 0x90, 0x00, 0x03, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, // MZ............
		0xB8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // .......@.......
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // ................
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, // ...............
		0x0E, 0x1F, 0xBA, 0x0E, 0x00, 0xB4, 0x09, 0xCD, 0x21, 0xB8, 0x01, 0x4C, 0xCD, 0x21, 0x54, 0x68, // .....!.L!Th
		0x69, 0x73, 0x20, 0x70, 0x72, 0x6F, 0x67, 0x72, 0x61, 0x6D, 0x20, 0x63, 0x61, 0x6E, 0x6E, 0x6F, // is program canno
		0x74, 0x20, 0x62, 0x65, 0x20, 0x72, 0x75, 0x6E, 0x20, 0x69, 0x6E, 0x20, 0x44, 0x4F, 0x53, 0x20, // t be run in DOS 
		0x6D, 0x6F, 0x64, 0x65, 0x2E, 0x0D, 0x0D, 0x0A, 0x24, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  // mode....$.......
	};

	const unsigned char pehdr[] = "PE\0"; // two trailing nulls, but one is implied by writing a character constant

	fwrite(mzstub, sizeof(mzstub), 1, lo.output);

	fwrite(pehdr, sizeof(pehdr), 1, lo.output);

	// Build base COFF header

	coffHeader ch = { 0 };
	ch.machine = MACHINE_TYPE_i386;
	ch.numSections = sectionData.size();
	ch.sizeOfOptionalHeader = sizeof(optHeader);
	ch.characteristics = ( // COFF_CHARACTERISTICS::
		IMAGE_FILE_RELOCS_STRIPPED |   // non-relocatable
		IMAGE_FILE_EXECUTABLE_IMAGE |  // .exe
		IMAGE_FILE_LINE_NUMS_STRIPPED |
		IMAGE_FILE_32BIT_MACHINE |
		IMAGE_FILE_DEBUG_STRIPPED
	);
	fwrite(&ch, sizeof(coffHeader), 1, lo.output);

	// Size of headers up to opt/image PE header inclusive

	uint32 fcurpos = (
		sizeof(mzstub) +
		sizeof(pehdr) +
		sizeof(coffHeader) +
		sizeof(optHeader)
	);

	// sectpos_acc: keep track of section file positions as we write out section headers
	uint32 sectpos_acc = fcurpos + (sectionData.size() * sizeof(sectionHeader));

	uint32 sectpos_pad = lo.fileAlign - (sectpos_acc % lo.fileAlign);
	if (sectpos_pad == lo.fileAlign) sectpos_pad = 0;

	sectpos_acc += sectpos_pad;

	// Build optional/image PE headers

	optHeader oh = { 0 };
	oh.s.magic = 0x10B;
	oh.s.addressOfEntryPoint = entry_rva; // RVA (i.e. not including base address offset) of entry point symbol

	if (info.sectionMemPos.find(".text") != info.sectionMemPos.end()) {
		oh.s.baseOfCode = info.sectionMemPos[".text"]; // COFF remnants, windows doesn't necessarily use this(?)
	} else {
		fprintf(stderr, "Warning: No .text segment in output\n");
		oh.s.baseOfCode = 0;
	}
	oh.s.baseOfData = 0; // lo.baseAddress;

	for (sectiondata_t::iterator i = sectionData.begin(), e = sectionData.end(); i != e; ++i) {

		uint32 len = i->second.length();
		uint32 align = info.sectionAlignmentsFile[i->first];

		if (i->first == ".text") {
			oh.s.sizeOfText += len + align;

		} else if (i->first == ".bss") {
			oh.s.sizeOfBSS += len + align;

		} else {
			oh.s.sizeOfData += len + align;
		}
	}

	oh.w.imageBase             = lo.baseAddress;
	oh.w.osVersionMajor        = 4;
	oh.w.osVersionMinor        = 0;
	oh.w.subsystemVersionMajor = 4;
	oh.w.subsystemVersionMinor = 0;
	oh.w.imageVersionMajor     = 1;
	oh.w.imageVersionMinor     = 0;
	oh.w.sizeOfImage           = info.totalImageLengthMem;
	oh.w.sizeOfHeaders         = sectpos_acc;
	oh.w.subsystem             = 0x3; // CONSOLE
	oh.w.alignFile             = lo.fileAlign;
	oh.w.alignSections         = lo.sectionAlign;
	oh.w.numRvaAndSizes        = 16;

	// Using mingw ld defaults here
	oh.w.sizeOfStackReserve    = 0x00200000; // 2MB
	oh.w.sizeOfStackCommit     = 0x00001000; // 1KB
	oh.w.sizeOfHeapReserve     = 0x00100000; // 1MB
	oh.w.sizeOfHeapCommit      = 0x00001000; // 1KB

	// -- symbol for entry point is set in struct lo
	// -- convert to new absolute location

	fwrite(&oh, sizeof(optHeader), 1, lo.output);

	// 
	//

	// Build section headers
	for (sectiondata_t::iterator i = sectionData.begin(), e = sectionData.end(); i != e; ++i) {

		uint32 len = i->second.length();

		sectionHeader sh = { 0 };

		sh.sizeOfRawData = len + info.sectionAlignmentsFile[i->first];
		sh.virtualSize = len;
		sh.characteristics = sectionCharacteristics[i->first];
		sh.virtualAddress = info.sectionMemPos[i->first];

		sh.fpRawData = sectpos_acc;		
		sectpos_acc += sh.sizeOfRawData;

		strncpy(sh.name, i->first.c_str(), 8);

		fwrite(&sh, sizeof(sectionHeader), 1, lo.output);
	}

	// write zeroes
	while (sectpos_pad-- > 0) {
		fputc('\x0', lo.output);
	}

	// echo all sections
	for (sectiondata_t::iterator i = sectionData.begin(), e = sectionData.end(); i != e; ++i) {

		fwrite(i->second.c_str(), i->second.length(), 1, lo.output);

		// pad section out to fileAlignment

		uint32 align = info.sectionAlignmentsFile[i->first];

		while (align--) { // any batch way of doing this? mmap/memset maybe
			fputc('\x0', lo.output);
		}

	}

	// No symbol/string table in output
	
	// done

	fclose(lo.output);

	return 0;
}


