/* mld/linker.h */ 

#ifndef MLD_LINKER_H
#define MLD_LINKER_H

#include <string>
#include <vector>
#include <set>

#include "file.h"

#include "cofflayout.h"

#include <cstdio>
#include <ctime>
#include <string>
#include <vector>
#include <map>

struct linkOptions {
	linkOptions() :
		baseAddress(0x00400000),
		sectionAlign(4096),
		fileAlign(512),
		verbose(true),
		output(NULL)
	{};

	unsigned int baseAddress;
	unsigned int sectionAlign;
	unsigned int fileAlign;
	
	bool verbose;

	FILE* output;

	std::string entry_func;
	std::vector<finfo> inputFiles;

	std::set<std::string> dropSections;

};


struct nice_relocation {
	uint16 type;
	uint32 va;
	std::string symbol;
};

struct nice_section {
	sectionHeader* raw;
	std::string name;
	std::string data;
	std::vector<nice_relocation> relocations;
};

struct nice_symbolinfo {
	std::string symbol;
	int fileid;
	std::string sectionName;
	uint32 sectionOffset;
};

typedef std::map<std::string, nice_section> section_map_t;
typedef std::map<std::string, std::string> sectiondata_t;
typedef std::map<std::string, uint32> sectioncharac_t;
typedef std::map<std::string, nice_symbolinfo> symbolmap_t;

struct coffInfo {
	coffHeader* ch;
	optHeader* oh;
	std::vector<std::string> sectionNumbers;
	section_map_t allSections;
	std::vector<coffSymbol*> unresolvedExternals, providedExternals, providedInternals;
	uint32* stringTableSizePtr;
	std::vector<char*> stringTableIndex;
};

typedef std::vector<coffInfo> vec_coffinfo_t;
typedef std::map<int, int> mii_t;
typedef std::map<std::string, mii_t> offset_t;

struct sectionOffsetInfo {
	sectioncharac_t sectionAlignmentsFile;
	sectioncharac_t sectionAlignmentsMem;
	sectioncharac_t sectionMemPos;
	uint32 totalImageLengthFile;
	uint32 totalImageLengthMem;

	sectionOffsetInfo(linkOptions& lo, sectiondata_t& sectionData);
};

// Function prototypes

int linker_process(linkOptions& lo); // linker.cpp

int linker_process_file(bool verbose, finfo& loadedFile, coffInfo& ret); // link_input.cpp

int linker_write_file(
	linkOptions& lo,
	sectiondata_t& sectionData,
	sectioncharac_t& sectionCharacteristics,
	uint32 entry_rva
); // link_output.cpp


#endif
