#!/bin/bash

CFLAGS="-Os -std=c89 -fno-asynchronous-unwind-tables -nostdlib -c"
CC="gcc"
LDFLAGS="-nostdlib -e_realmain"
LD="ld"
MLDFLAGS="--drop-section .drectve -e _realmain"
MLD="../Debug/mld.exe"

tidyup() {
	rm *.o
	[ -f test__ld.exe ] && rm test__ld.exe
	[ -f test__mld.exe ] && rm test__mld.exe
}

dotest() {	
	tidyup
	
	echo -n "Testing... "
	
	local ldcmd="$LD $LDFLAGS "
	local mldcmd="$MLD $MLDFLAGS "

	for i in "$@" ; do
		BASENAME=$(basename "$i" .c)
		
		# Compile each input file
		if ! $CC $CFLAGS "$i" -o $BASENAME.o ; then
			echo "Compilation failed! ($?)"
			exit 1
		fi
		
		ldcmd="$ldcmd $BASENAME.o "
		mldcmd="$mldcmd -i $BASENAME.o "
	done
	
	# Link with GNU ld and MLD
	eval "$ldcmd -o test__ld.exe"
	eval "$mldcmd -o test__mld.exe"
	
	# Compare output
	./test__ld.exe
	LDRET=$?
	./test__mld.exe
	MLDRET=$?
	
	if [ "$LDRET" == "$MLDRET" ] ; then
		echo "Passed! ($MLDRET)"
	
	else
		echo "FAILED ($LDRET != $MLDRET)";
		exit 1
	fi
}

main() {
	dotest test01*
	dotest test02*
	dotest test03*
	dotest test04*
	tidyup
}

main
