#!/bin/bash

CFLAGS="-Os -std=c89 -fno-asynchronous-unwind-tables -nostdlib -c"
CC="i686-w64-mingw32-gcc"
LDFLAGS="-nostdlib -e_realmain"
LD="i686-w64-mingw32-ld"
MLDFLAGS="--drop-section .drectve -e _realmain --compress-rdata"
MLD="../Debug/mld.exe"

tidyup() {
	rm *.o
	[ -f test__ld.exe ] && rm test__ld.exe
	[ -f test__mld.exe ] && rm test__mld.exe
}

extra_set() {
	EXTRA="$@"
}

dotest() {	
	tidyup
	
	echo -n "Testing... "
	
	local ldcmd="$LD $LDFLAGS "
	local mldcmd="$MLD $MLDFLAGS "
	local nameprefix="$(echo "$1" | cut -d_ -f1)"
	
	for i in "$@" ; do
		BASENAME=$(basename "$i" .c)
		
		# Compile each input file
		if ! $CC $CFLAGS "$i" -o $BASENAME.o ; then
			echo "Compilation failed! ($?)"
			exit 1
		fi
		
		ldcmd="$ldcmd $BASENAME.o "
		mldcmd="$mldcmd -i $BASENAME.o "
	done
	
	for i in $EXTRA ; do
		ldcmd="$ldcmd $i "
		mldcmd="$mldcmd -i $i "
	done
	
	# Link with GNU ld and MLD
	eval "$ldcmd -o ${nameprefix}_output_ld.exe"
	eval "$mldcmd -o ${nameprefix}_output_mld.exe"
	
	# Compare output
	"./${nameprefix}_output_ld.exe"
	LDRET=$?
	"./${nameprefix}_output_mld.exe"
	MLDRET=$?
	
	if [ "$LDRET" == "$MLDRET" ] ; then
		echo "Passed! ($MLDRET)"
	
	else
		echo "FAILED ($LDRET != $MLDRET)";
		exit 1
	fi
	
	extra_set
}

main() {
	dotest test01*.c
	dotest test02*.c
	dotest test03*.c
	dotest test04*.c
	# 05: win32
	dotest test06*.c
	
	extra_set libmsvcrt.a
	dotest test07*.c
	
	extra_set libkernel32.a
	dotest test08*.c
	
	tidyup
}

main
read -p "Press any key to continue..."
