/* mld/defs.h */

#ifndef MLD_COFFLAYOUT_H
#define MLD_COFFLAYOUT_H

#if defined(_MSC_VER)
	#define PACK( __Declaration__ ) __Declaration__
	// #define PACK( __Declaration__ ) __pragma( pack(push, 1) ) __Declaration__ __pragma( pack(pop) )
#elif defined(__GNUC__)
	#define PACK( __Declaration__ ) __Declaration__ __attribute__((__packed__))
#else
	#error "The PACK macro is not implemented for this compiler"
#endif

typedef signed short sint16;
typedef unsigned short uint16;
typedef unsigned int uint32;
typedef unsigned char uchar;

PACK(struct dosHeader{
	uchar magic[2];
	uchar unused[0x3c - 2];
	uint32 pe_offset;
});

PACK(struct coffHeader{
	uint16 machine;
	uint16 numSections;
	uint32 timeStamp;
	uint32 ptrSymbolTable; /* zero/deprecated */
	uint32 numSymbols;     /* zero/deprecated */
	uint16 sizeOfOptionalHeader;
	uint16 characteristics;
});

PACK(struct dataDirectory{
	uint32 virtualAddr;
	uint32 size;
});

PACK(struct optionalHeaderStandard{
	uint16 magic;
	uchar linkerVersionMajor;
	uchar linkerVersionMinor;
	uint32 sizeOfText;
	uint32 sizeOfData;
	uint32 sizeOfBSS;
	uint32 addressOfEntryPoint;
	uint32 baseOfCode;
	uint32 baseOfData; /* absent in PE32+ */
});

PACK(struct optionalHeaderWin32{
	uint32 imageBase;
	uint32 alignSections;
	uint32 alignFile;
	uint16 osVersionMajor; /* minimum OS version */
	uint16 osVersionMinor;
	uint16 imageVersionMajor;
	uint16 imageVersionMinor;
	uint16 subsystemVersionMajor;
	uint16 subsystemVersionMinor;
	uint32 win32Version; /* must be zero */
	uint32 sizeOfImage;
	uint32 sizeOfHeaders;
	uint32 checksum; /* optional */
	uint16 subsystem;
	uint16 dllCharacteristics;
	uint32 sizeOfStackReserve;
	uint32 sizeOfStackCommit;
	uint32 sizeOfHeapReserve;
	uint32 sizeOfHeapCommit;
	uint32 loaderFlags; /* must be zero */
	uint32 numRvaAndSizes;
});

PACK( struct optionalHeaderDirectory{
	dataDirectory exportTable;
	dataDirectory importTable;
	dataDirectory resourceTable;
	dataDirectory exceptionTable;
	dataDirectory certificateTable;
	dataDirectory baseRelocationTable;
	dataDirectory debug;
	dataDirectory architecture; /* must be zero */
	dataDirectory globalPtr;
	dataDirectory tlsTable;
	dataDirectory loadConfigTable;
	dataDirectory boundImport;
	dataDirectory importAddressTable;
	dataDirectory delayImportDescriptor;
	dataDirectory CLRRuntimeHeader;
	dataDirectory reserved; /* must be zero */
});

PACK(struct optHeader{
	optionalHeaderStandard s;
	optionalHeaderWin32 w;
	optionalHeaderDirectory d;
});

PACK(struct sectionHeader{
	char name[8];
	uint32 virtualSize;
	uint32 virtualAddress;
	uint32 sizeOfRawData;
	uint32 fpRawData;
	uint32 fpRelocations;
	uint32 fpLineNumbers; /* zero, deprecated */ 
	uint16 numRelocations;
	uint16 numLineNumbers; /* zero, deprecated */ 
	uint32 characteristics;
});

PACK(struct relocation{
	uint32 virtualAddress;
	uint32 symbolTableIndex;
	uint16 type;
});

#define SIZEOF_RELOCATION 10

PACK(struct coffSymbol{
	union {
		char shortName[8];
		struct {
			uint32 zero;
			uint32 stringTableOffset;
		} offset;
	} name;

	uint32 value;
	sint16 sectionNumber; /* signed */ 
	uint16 type;
	uchar storageClass;
	uchar numAuxSymbols;
});

#define SIZEOF_COFFSYMBOL 18

#define MACHINE_TYPE_i386 0x14c

#define RELOCATION_TYPE_DIR32 0x6
#define RELOCATION_TYPE_DIR32NB 0x7
#define RELOCATION_TYPE_REL32 0x14

#define SUBSYSTEM_TYPE_WINGUI 0x02
#define SUBSYSTEM_TYPE_WINCLI 0x03

enum COFF_CHARACTERISTICS {
	IMAGE_FILE_RELOCS_STRIPPED = 0x0001,
	IMAGE_FILE_EXECUTABLE_IMAGE = 0x0002,
	IMAGE_FILE_LINE_NUMS_STRIPPED = 0x0004,
	IMAGE_FILE_LOCAL_SYMS_STRIPPED = 0x0008,
	IMAGE_FILE_AGGRESSIVE_WS_TRIM = 0x0010,
	IMAGE_FILE_LARGE_ADDRESS_AWARE = 0x0020,
	IMAGE_FILE_BYTES_REVERSED_LO = 0x0080,
	IMAGE_FILE_32BIT_MACHINE = 0x0100,
	IMAGE_FILE_DEBUG_STRIPPED = 0x0200,
	IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP = 0x0400,
	IMAGE_FILE_NET_RUN_FROM_SWAP = 0x0800,
	IMAGE_FILE_SYSTEM = 0x1000,
	IMAGE_FILE_DLL = 0x2000,
	IMAGE_FILE_UP_SYSTEM_ONLY = 0x4000,
	IMAGE_FILE_BYTES_REVERSED_HI = 0x8000
};

enum SECTION_CHARACTERISTICS {
	IMAGE_SCN_CNT_CODE = 0x20,
	IMAGE_SCN_CNT_INITIALIZED_DATA = 0x40,
	IMAGE_SCN_CNT_UNINITIALIZED_DATA = 0x80,

	IMAGE_SCN_MEM_EXECUTE = 0x20000000,
	IMAGE_SCN_MEM_READ = 0x40000000,
	IMAGE_SCN_MEM_WRITE = 0x80000000
};

#define ARCHIVE_HEADER "!<arch>\n"
#define ARCHIVE_HEADER_LEN 8

PACK(struct archiveMember {
	char name[16];
	char date[12];
	char uid[6];
	char gid[6];
	char mode[8];
	char size[10];
	char endOfHeader[2];
});

#endif
