/* link_archive.cpp */

#include "linker.h"

bool linker_separate_archive(finfo& input, std::vector<finfo>& ret) {

	size_t curPos = ARCHIVE_HEADER_LEN;

	ret.clear();
	
	bool is_first = true;

	for (;;) {

		// Is there room in input stream for another archiveMember (or are we at EOF)?

		if (curPos + sizeof(archiveMember) >= input.data.length()) {
			return true;
		}

		// Read header

		archiveMember am = { 0 };
		memcpy(&am, input.data.c_str() + curPos, sizeof(archiveMember));

		curPos += sizeof(archiveMember);

		// Is there room in input stream for declared data size?

		size_t dataSize = atoi(am.size);
		if (curPos + dataSize > input.data.length()) {
			fprintf(stderr, "FATAL: Archive member declared %d size but only %d bytes remaining in file\n",
				dataSize,
				(input.data.length() - curPos)
			);
			return false; // corrupt
		}

		// Read data

		finfo nested;

		char buff[32] = { 0 };
		sprintf(buff, "%%0x%x", curPos);
		nested.filename = input.filename + buff;

		nested.data = input.data.substr(curPos, dataSize);
		curPos += dataSize;

		// Stash (except the first, which is metadata)

		if (is_first) {
			is_first = false;
		} else {
			ret.push_back(nested);
		}

		// Round curPos up to nearest 2

		if (curPos % 2) {
			curPos++;
		}

		// fprintf(stderr, "Got %.*s size 0x%X next pos 0x%X\n", sizeof(am.name), am.name, dataSize, curPos);
	}

	// UNREACHED
}

bool linker_file_is_archive(finfo& f) {
	return (f.data.substr(0, ARCHIVE_HEADER_LEN) == ARCHIVE_HEADER);
}

bool linker_separate_archives(linkOptions& lo) {

	// Expand any passed-in .a files into their constituent .obj files.

	for (int i = 0; i != lo.inputFiles.size() /* recheck every loop */; ++i) {

		if (linker_file_is_archive(lo.inputFiles[i])) {

			// It's an archive

			std::vector<finfo> separated;
			if (!linker_separate_archive(lo.inputFiles[i], separated)) {
				fprintf(stderr, "FATAL: Failed to parse archive file '%s'\n", lo.inputFiles[i].filename.c_str());
				return false;
			}

			// Append

			lo.inputFiles.insert(lo.inputFiles.end(), separated.begin(), separated.end());

			// Free the input archive, and shift back one index

			lo.inputFiles.erase(lo.inputFiles.begin() + i);
			i--;

		}

	}

	return true;

}
