/* mld/main.cpp */

#include <cstdio>
#include <cstring> // strcmp
#include <cstdlib> // atoi

#include "file.h"
#include "linker.h"

int usage() {
	fputs(
		"Usage: mld [options]\n"
		"\n"
		"   --align-file {number}      Set section alignment in file (default 512)\n"
		"   --align-section {number}   Set section alignment in memory (default 4096)\n"
		"   --drop-section {name}      Do not include section in output\n"
		"   -e {symbol name}           Set entry point (default 'main')\n"
		"   -h                         Display this message\n"
		"   -i {filename}              Add input file to link\n"
		"   -mconsole                  Use CONSOLE subsystem (default)\n"
		"   -mwindows                  Use WINDOWS subsystem\n"
		"   -o {filename}              Set output file, - for stdout\n"
		,
		stderr
	);
	return -1;
}

#define REQUIRE_FOLLOWING_ARG { if (i + 1 == argc) return usage(); }

int main(int argc, char** argv) {

	/*

	Test args (Alt+Shift+5)	
	-v --drop-section .drectve -e _realmain -i ..\test\test03_text.o -i ..\test\test03_data.o -o ..\test\ide.exe

	*/

	linkOptions lo;
	lo.entry_func = "main";

	for (int i = 1; i < argc; ++i) {

		if (!strcmp(argv[i], "--align-file")) {
			REQUIRE_FOLLOWING_ARG;
			lo.fileAlign = atoi(argv[++i]);

		} else if (!strcmp(argv[i], "--align-section")) {
			REQUIRE_FOLLOWING_ARG;
			lo.sectionAlign = atoi(argv[++i]);

		} else if (!strcmp(argv[i], "--drop-section")) {
			REQUIRE_FOLLOWING_ARG;
			lo.dropSections.insert(std::string(argv[++i]));

		} else if (!strcmp(argv[i], "-e")) {
			REQUIRE_FOLLOWING_ARG;
			lo.entry_func = std::string(argv[++i]);

		} else if (!strcmp(argv[i], "-h")) {
			return usage();

		} else if (!strcmp(argv[i], "-mwindows")) {
			lo.isWindowsSubsystem = true;

		} else if (!strcmp(argv[i], "-mconsole")) {
			lo.isWindowsSubsystem = false;

		} else if (!strcmp(argv[i], "-i")) {
			REQUIRE_FOLLOWING_ARG;
			finfo loadedFile;
			if (!loadFile(argv[++i], loadedFile)) {
				fprintf(stderr, "Error loading file '%s'\n", argv[i]);
				return 1;
			}
			lo.inputFiles.push_back(loadedFile);

		} else if (!strcmp(argv[i], "-o")) {
			REQUIRE_FOLLOWING_ARG;
			lo.output = (strcmp(argv[++i], "-")) ? fopen(argv[i], "wb") : stdout;
			if (! lo.output) {
				fprintf(stderr, "Error opening output file '%s'\n", argv[i]);
				return 1;
			}

		} else {
			fprintf(stderr, "Unrecognised command '%s'\n", argv[i]);
			return usage();
		}

	}

	if (!lo.inputFiles.size()) {
		return usage();
	}

	if (!linker_separate_archives(lo)) {
		return 1; // failure
	}

	return linker_process(lo);

}
